import pytest

from spilo.base_client import BaseClient


class Client(BaseClient):

    async def send(self, data):
        pass

    async def close(self):
        pass


@pytest.fixture
def client():
    c = Client()
    yield c




import pytest
from spilo.channel import Channel


def test_channel_instance():
    channel = Channel("test_channel")
    isinstance(channel.clients, set)
    isinstance(channel.dict_clients, dict)


def test_adding_new_client_in_channel(client):
    channel = Channel("test_channel")
    channel.add_client(client)

    assert len(channel.clients) == 1
    for client_1 in channel.clients:
        assert client_1 is client


@pytest.mark.asyncio
async def test_remove_client_from_channel(client):
    channel = Channel("test_channel")
    channel.add_client(client)

    await channel.remove_client(client)
    assert len(channel.clients) == 0
    assert channel.dict_clients.get(client.client_id) is None




from abc import ABC, abstractmethod
from typing import AsyncIterable


class BasePubSub(ABC):
    """
    This is abstract base class defines interface
    which should be implemented by any child class.
    """

    @abstractmethod
    async def connect(self):
        """
        Method for connecting pub sub backend.
        """

    @abstractmethod
    async def publish(self, channel_name: str, data):
        """
        Method for publishing messages.
        """

    @abstractmethod
    async def listen(self, channel_name: str) -> AsyncIterable:
        """
        Method for listening incoming messages.
        """




import json

import aioredis
from aioredis import Redis
from aioredis.client import PubSub

from .base_pubsub import BasePubSub


class RedisPubSub(BasePubSub):

    _singleton = None
    
    def __new__(cls, url="redis://localhost:6379/0", redis_options=None, *args, **kwargs):
        if not cls._singleton:
            cls._singleton = super(RedisPubSub, cls).__new__(cls, *args, **kwargs)
        return cls._singleton

    def __init__(self, url="redis://localhost:6379/0", redis_options=None):
        self.redis_url = url
        self.redis_options = redis_options or {}
        self.redis: Redis | None = None
        self.pubsub: PubSub | None = None

    def connect(self):
        self.redis = aioredis.Redis.from_url(self.redis_url, decode_responses=True, **self.redis_options)
        self.pubsub = self.redis.pubsub(ignore_subscribe_messages=True)

    async def publish(self, channel_name: str, data):
        return await self.redis.publish(
            channel_name, json.dumps(data)
        )

    async def listen(self, channel_name: str):
        await self.pubsub.subscribe(channel_name)
        async for message in self.pubsub.listen():
            print(message)
            yield message
        await self.pubsub.unsubscribe(channel_name)



import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass



@dataclass(kw_only=True)
class BaseClient(ABC):
    
    client_id: uuid.UUID = uuid.uuid4()

    def __hash__(self):
        return self.client_id.int

    @abstractmethod
    async def send(self, data: bytes | str):
        pass

    @abstractmethod
    async def close(self):
        pass




# from dataclasses import dataclass
#
# from spilo.base_client import BaseClient
# from spilo.channel import Channel
#
#
# def test_overiding_id_field_in_base_client():
#
#     @dataclass
#     class Client(BaseClient):
#
#         client_id: str
#         __hash__ = BaseClient.__hash__
#
#         async def send(self, data):
#             pass
#
#         async def close(self):
#             pass
#
#     client = Client("test_client_id")
#     channel = Channel("test_channel")
#
#     assert client.client_id == "test_client_id"
#     assert client.__hash__ is not None
#
#     channel.add_client(client)
#
#     assert len(channel.clients) == 1
#



from spilo.redis_pubsub import RedisPubSub


def test_singleton_class():
    cls1 = RedisPubSub(url="redis://localhost:6379/0")
    cls2 = RedisPubSub(url="redis://localhost:6379/0")
    assert cls1 is cls2




from typing import Set, Dict, Any, Tuple

from .base_client import BaseClient
from .base_pubsub import BasePubSub


class Channel:
    """
    Class for handling clients adding new clients, removing clients in one specific channel.
    :param channel_name: The channel name where clients|client will be subscribed.
    clients will send and receive messages through channel_name.
    """

    channel_cache = {}

    def __init__(self, channel_name: str, pubsub_manager: BasePubSub):
        self.channel_name = channel_name
        self.clients: Set[BaseClient] = set()
        self.dict_clients: Dict[Any, BaseClient] = {}
        self.pubsub_manager = pubsub_manager

    @classmethod
    def get(cls, channel_name: str, pubsub_manager: BasePubSub) -> Tuple["Channel", bool]:
        """
        Class method for getting channel class if channel class does not exist
        method will create new one and will return from function.
        """
        if channel_name in cls.channel_cache:
            return cls.channel_cache[channel_name], True
        channel = cls(channel_name, pubsub_manager)
        cls.channel_cache[channel_name] = channel
        return channel, False

    def add_client(self, client: BaseClient) -> None:
        """
        Method for adding ws clients to channel.
        """
        self.clients.add(client)
        self.dict_clients[client.client_id] = client

    async def remove_client(self, client: BaseClient):
        """
        Method for removing client from channel.
        """
        if client in self.clients:
            await client.close()
            self.clients.remove(client)
            self.dict_clients.pop(client.client_id, None)

    async def receiver(self):
        async for raw in self.pubsub_manager.listen(self.channel_name):
            if raw:
                if raw["data"] == "STOP" and len(self.clients) == 0:
                    break
                if raw["channel"] != self.channel_name:
                    try:
                        await self.dict_clients[raw["channel"]].send(
                            raw["data"]
                        )
                    except KeyError:
                        # TODO LOG SOMETHING
                        pass
                else:
                    for client in self.clients:
                        await client.send(str(raw["data"]))

    async def publish(self, data):
        await self.pubsub_manager.publish(self.channel_name, data)
